/*
	ncbricks - ncurses based brick breaker
	Copyright (C) 2017-2018  Aron Barath

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
	02110-1301  USA
*/

#include "ncbricks.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void brick_show_all()
{
	size_t i;

	for(i=0;i<g_bricks.size;++i)
	{
		brick_show((struct brick*)g_bricks.pointers[i]);
	}
}

void brick_hide_all()
{
	size_t i;

	for(i=0;i<g_bricks.size;++i)
	{
		brick_hide((struct brick*)g_bricks.pointers[i]);
	}
}

static const char brick_inner   = '=';
static const char brick_left[]  = { '#', '(', '[', '{' };
static const char brick_right[] = { '#', ')', ']', '}' };

/* this is ugly as hell, but we must ensure some numbers to match */
struct check_BRICK_MAXHP
{
	int static_assert_field
		:
		((BRICK_MAXHP+1) == sizeof(brick_left))
		&&
		((BRICK_MAXHP+1) == sizeof(brick_right))
		;
};

void brick_show(const struct brick* const brick)
{
	const int cpair = brick->cpair;

	if(g_use_colors)
	{
		wattron(g_game_win, COLOR_PAIR(cpair));
	}

	wmove(g_game_win, g_board_top+brick->pos.pos.y, g_board_left+brick->pos.pos.x);
	waddch(g_game_win, brick_left[(unsigned char)brick->hitpts]);
	{
		size_t i;

		for(i=2;i<BRICK_WIDTH;++i)
		{
			waddch(g_game_win, brick_inner);
		}
	}
	waddch(g_game_win, brick_right[(unsigned char)brick->hitpts]);

	if(g_use_colors)
	{
		wattroff(g_game_win, COLOR_PAIR(cpair));
	}
}

void brick_hide(const struct brick* const brick)
{
	size_t i;

	wmove(g_game_win, g_board_top+brick->pos.pos.y, g_board_left+brick->pos.pos.x);

	for(i=0;i<BRICK_WIDTH;++i)
	{
		waddch(g_game_win, ' ');
	}
}

static bool parse_brick_pair(const int x, const int y, const char hp, const char charm)
{
	struct brick* brick;

	if(' '==hp)
	{
		return TRUE;
	}

	if(!('0'<=hp && hp<=('0'+BRICK_MAXHP)))
	{
		show_message("Invalid hit point in level data.");
		return FALSE;
	}

	if(NULL==(brick=(struct brick*)malloc(sizeof(struct brick))))
	{
		show_message("No memory to allocate brick.");
		return FALSE;
	}

	memset(brick, 0, sizeof(struct brick));

	brick->hitpts = hp - '0';
	brick->charm  = charm_decode(charm);

	if(0==brick->hitpts)
	{
		brick->cpair = CPAIR_BRICK_U;
		++g_num_unbreakable;
	}
	else
	{
		brick->cpair = CPAIR_BRICK_BASE + (rand() % NUM_BRICK_COLORS);
	}

	brick->pos.pos.x = x;
	brick->pos.pos.y = y;
	brick->pos.end.x = x + BRICK_WIDTH - 1;
	brick->pos.end.y = y;

	brick_show(brick);
	ptrarr_add(&g_bricks, brick);

	return TRUE;
}

#define MAX_BRICK_Y    (GAME_HEIGHT-8)

bool brick_load_data(int level)
{
	bool res = FALSE;
	char filename[] = LEVELS_PATH "99";
	char line[BRICKS_PER_ROW*2+2];
	FILE* file;
	char* end;
	int y;

	if(0>level)
	{
		show_message("Level index is too low.");
		return FALSE;
	}
	// TODO: stringify!
	if(99<level)
	{
		show_message("Level index is too high.");
		return FALSE;
	}

	sprintf(filename+sizeof(LEVELS_PATH)-1, "%i", level);

	if(NULL==(file=fopen(filename, "rb")))
	{
		return FALSE;
	}

	if(NULL==fgets(line, sizeof(line), file))
	{
		show_message("Level data is corrupted (1).");
		goto out;
	}

	g_num_unbreakable = 0;

	/* the first line contains a decimal number, this number adjusts */
	/* the starting Y coordinate of the bricks -- but note, this is */
	/* equivalent to this: "the number of empty rows above the first row */
	/* that contains bricks", since ncurses uses 0-based indices */

	y = (int)strtol(line, &end, 10);
	if(!end || !(0==*end || '\n'==*end) || 0>y)
	{
		show_message("Level data is corrupted (2).");
		goto out;
	}
	if(MAX_BRICK_Y<y)
	{
		show_message("Level is too high (1).");
		goto out;
	}

	/* the next lines will contain row data, */
	/* row data contains BRICKS_PER_ROW character pair */
	/* (so-called "brick pair") */

	while(NULL!=fgets(line, sizeof(line), file) && 0!=line[0])
	{
		size_t i;

		if(MAX_BRICK_Y<y)
		{
			show_message("Level is too high (2).");
			goto out;
		}

		if(strlen(line)!=sizeof(line)-1)
		{
			show_message("Level data is corrupted (3).");
			goto out;
		}

		/* parse all brick pair in the current row */

		for(i=0;i<BRICKS_PER_ROW;++i)
		{
			if(!parse_brick_pair(i*BRICK_WIDTH, y, line[2*i+0], line[2*i+1]))
			{
				goto out;
			}
		}

		/* we increment Y here, because we go to the next row */
		/* in the iteration of the loop */

		++y;
	}

	// TODO: did we load at least one line?

	res = TRUE;
out:
	fclose(file);
	return res;
}

void brick_delete_all()
{
	while(0<g_bricks.size)
	{
		struct brick* const brick = (struct brick*)g_bricks.pointers[--g_bricks.size];
		brick_hide(brick);
		free(brick);
	}
}
