/*
	ncbricks - ncurses based brick breaker
	Copyright (C) 2017-2018  Aron Barath

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
	02110-1301  USA
*/

#ifndef NCBRICKS_H
#define NCBRICKS_H

#include "config.h"
#include "vec2.h"
#include "ptrarr.h"

#define GAME_WIDTH       80
#define GAME_HEIGHT      24
#define BRICK_WIDTH      6
#define LOGICAL_SCALE    64 /* scale of the logical coordinate system */
#define BALL_VELOCITY    64
#define BALL_ATOMICTIME  15 /* number of seconds the atomic ball lasts */
#define BALL_SLOWTIME    15 /* number of seconds the slow ball lasts */
#define MAX_LIFE         9999
#define MAX_AMMO         999999
#define MAX_BALLS        5 /* maximal balls on the board (split charm) */
#define INITIAL_AMMO     5 /* number of ammo the player initially gets */

#define BRICK_MAXHP      3
#define BRICKS_PER_ROW   ((GAME_WIDTH-2)/BRICK_WIDTH)

#define PADDLE_WIDTH     9
#define PADDLE_MINW      3
#define PADDLE_MAXW      25
#define PADDLE_TINY      1 /* tiny step */
#define PADDLE_SMALL     3 /* small step */
#define PADDLE_LARGE     6 /* large step */

#define CPAIR_TEXT       1
#define CPAIR_BALL       2
#define CPAIR_BALL2      3 /* atomic ball */
#define CPAIR_PADDLE     4
#define CPAIR_CHARM      5
#define CPAIR_MISSILE    6
#define CPAIR_STATUS     9
#define CPAIR_BRICK_U    10
#define CPAIR_BRICK_1    11
#define CPAIR_BRICK_2    12
#define CPAIR_BRICK_3    13
#define CPAIR_BRICK_4    14
#define CPAIR_BRICK_5    15
#define CPAIR_BRICK_6    16
#define CPAIR_BRICK_BASE CPAIR_BRICK_1
#define NUM_BRICK_COLORS 6

#define LEVELS_PATH   "levels/"

#define SCORE_BRICK      100  /* base points to earn when hitting a brick */
#define SCORE_LEVEL      500  /* score to earn when level is cleared */
#define SCORE_EXTRA      1000 /* amount of score to earn when catch the charm */

#define AMMO_EXTRA1      5 /* amount of ammo to gain from charm type 1 */
#define AMMO_EXTRA2      9 /* amount of ammo to gain from charm type 2 */

struct ball
{
	struct vec2 dpos; /* position in display space */
	struct vec2 lpos; /* position in logical space */
	struct vec2 velo; /* velocity in the logical space */
	int         hits; /* number of hits since last contact with the paddle */
};

struct paddle
{
	int          x;     /* horizontal position in display space */
	int          width; /* width in display space */
	struct rect2 lrect; /* paddle rectangle in logical space */
};

struct brick
{
	struct rect2 pos;    /* position in display space */
	int          cpair;  /* display color (a CPAIR value) */
	char         hitpts; /* number of hits needed to destroy the brick (-1 is unbreakable) */
	char         charm;  /* the charm the brick will release when broken */
};

struct charm
{
	struct vec2 pos;    /* position in display space */
	char        charm;  /* charm symbol */

	void (*effect)();
};

struct charmdesc
{
	const char* desc;
	char        symbol;

	void (*effect)();
};

struct missile
{
	struct vec2 pos; /* position in display space */
};

extern WINDOW* g_stat_win; /* the window the status is in */
extern WINDOW* g_game_win; /* the window to play in */
extern struct paddle g_paddle;
extern struct ptrarr g_balls;
extern struct ptrarr g_bricks;
extern struct ptrarr g_charms;
extern struct ptrarr g_missiles;
extern bool g_use_colors;
extern int g_board_top; /* top offset when displaying stuff */
extern int g_board_left; /* left offset when displaying stuff */
extern int g_board_width;
extern int g_board_height;
extern int g_paddle_y;
extern size_t g_num_unbreakable; /* number of unbreakable bricks */
extern int g_level;
extern unsigned int g_life;
extern unsigned int g_ammo;
extern unsigned long long g_score;
extern int g_backdef;
extern int g_atomicball;
extern int g_slowball;

void invalidate_status();
void invalidate_board();
void show_message(const char* const msg);
void enable_atomic_ball();
void enable_slow_ball();

/* status helper functions */
void add_score(unsigned int n);
void add_ammo(unsigned int n);
void add_life(int n);

/* animate */
void animate_all(bool anim_balls);

/* paddle */
void paddle_reset(struct paddle* const paddle);
void paddle_show(const struct paddle* const paddle);
void paddle_hide(const struct paddle* const paddle);
void paddle_move(struct paddle* const paddle, int amount);
void paddle_enlarge(struct paddle* const paddle, int size);

/* ball */
void ball_show(const struct ball* const ball);
void ball_hide(const struct ball* const ball);
void ball_show_all();
void ball_hide_all();
void ball_delete_all();
void ball_spawn();
void ball_split_all();
void ball_reflect(struct ball* const ball, const struct rect2* const rect);

/* bricks */
void brick_show(const struct brick* const brick);
void brick_hide(const struct brick* const brick);
void brick_show_all();
void brick_hide_all();
void brick_delete_all();
bool brick_load_data(int level);

/* missile */
void missile_show(const struct missile* const missile);
void missile_hide(const struct missile* const missile);
void missile_show_all();
void missile_hide_all();
void missile_delete_all();
void missile_spawn(const struct paddle* const paddle);
void missile_spawn_apocalypse();

/* charm */
void charm_show(const struct charm* const charm);
void charm_hide(const struct charm* const charm);
void charm_show_all();
void charm_hide_all();
void charm_delete_all();
void charm_spawn(const struct brick* const brick, const char typ);
char charm_decode(const char charm);

#endif /* NCBRICKS_H */
