/*
	ncbricks - ncurses based brick breaker
	Copyright (C) 2017-2018  Aron Barath

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
	02110-1301  USA
*/

#include "vec2.h"
#include <stdint.h>
#include <stdlib.h>

static int64_t sqrt64(int64_t n0)
{
	uint64_t n=n0, m, y, b;

	if(n0<=0) return 0;

	m = UINT64_C(0x4000000000000000);
	y = UINT64_C(0);

	while(m != 0ul)
	{
		b = y | m;
		y = y >> 1;
		if(n >= b)
		{
			n = n - b;
			y = y | m;
		}
		m = m >> 2;
	}

	return y;
}

static int min_int(const int a, const int b)
{
	if(a<b) { return a; } return b;
}

static int max_int(const int a, const int b)
{
	if(a<b) { return b; } return a;
}

int vec2_length(const struct vec2* const vec)
{
	return (int)sqrt64(((long long)vec->x)*vec->x + ((long long)vec->y)*vec->y);
}

int vec2_distance(const struct vec2* const p1, const struct vec2* const p2)
{
	const long long a = ((long long)p1->x) - p2->x;
	const long long b = ((long long)p1->y) - p2->y;
	return (int)sqrt64(a*a + b*b);
}

int vec2_L1norm(const struct vec2* const p1, const struct vec2* const p2)
{
	int x = p1->x - p2->x;
	int y = p1->y - p2->y;
	if(0>x) { x = -x; }
	if(0>y) { y = -y; }
	return x + y;
}

void vec2_smul(struct vec2* const dst, const struct vec2* const vec, int scalar)
{
	dst->x = vec->x * scalar;
	dst->y = vec->y * scalar;
}

void vec2_sdiv(struct vec2* const dst, const struct vec2* const vec, int scalar)
{
	dst->x = vec->x / scalar;
	dst->y = vec->y / scalar;
}

void vec2_smuldiv(struct vec2* const dst, const struct vec2* const vec, int s_mul, int s_div)
{
	dst->x = (vec->x * s_mul) / s_div;
	dst->y = (vec->y * s_mul) / s_div;
}

void vec2_add(struct vec2* const dst, const struct vec2* const v1, const struct vec2* const v2)
{
	dst->x = v1->x + v2->x;
	dst->y = v1->y + v2->y;
}

void vec2_sub(struct vec2* const dst, const struct vec2* const v1, const struct vec2* const v2)
{
	dst->x = v1->x - v2->x;
	dst->y = v1->y - v2->y;
}

void vec2_random(struct vec2* const dst, const int magnitude)
{
	const long long x  = ((((long long)rand()) * magnitude) / (RAND_MAX/2)) - magnitude;
	const long long y  = ((((long long)rand()) * magnitude) / (RAND_MAX/2)) - magnitude;
	const long long l2 = sqrt64(x*x + y*y);

	dst->x = (int)((x * magnitude) / l2);
	dst->y = (int)((y * magnitude) / l2);
}

static int vec2_intersect_impl(struct vec2* const point,
	const int x1, const int y1,
	const int x2, const int y2,
	const int x3, const int y3,
	const int x4, const int y4)
{
	const int d = (x1-x2) * (y3-y4) - (y1-y2) * (x3-x4);

	if(0==d)
	{
		return -1;
	}
	else
	{
		const int pre  = (x1*y2 - y1*x2);
		const int post = (x3*y4 - y3*x4);
		const int x    = (pre*(x3-x4) - (x1-x2)*post) / d;
		const int y    = (pre*(y3-y4) - (y1-y2)*post) / d;

		/* check if the x and y coordinates are within both lines */
		if(x<min_int(x1, x2) || x>max_int(x1, x2) ||
			x<min_int(x3, x4) || x>max_int(x3, x4))
		{
			return -1;
		}
		if(y<min_int(y1, y2) || y>max_int(y1, y2) ||
			y<min_int(y3, y4) || y>max_int(y3, y4))
		{
			return -1;
		}

		point->x = x;
		point->y = y;

		return 0;
	}
}

int vec2_intersect(struct vec2* const point, const struct vec2* const line1s, const struct vec2* const line1e, const struct vec2* const line2s, const struct vec2* const line2e)
{
	return vec2_intersect_impl(point,
		line1s->x, line1s->y,
		line1e->x, line1e->y,
		line2s->x, line2s->y,
		line2e->x, line2e->y);
}

int vec2_intersect_vr(struct vec2* const point, const struct vec2* const line1s, const struct vec2* const line1e, const int line2sx, const int line2sy, const int line2ex, const int line2ey)
{
	return vec2_intersect_impl(point,
		line1s->x, line1s->y,
		line1e->x, line1e->y,
		line2sx, line2sy,
		line2ex, line2ey);
}

static void swap_int(int* const p1, int* const p2)
{
	const int tmp = *p1;
	*p1 = *p2;
	*p2 = tmp;
}

void rect2_normalize(struct rect2* const rect)
{
	if(rect->pos.x>rect->end.x)
	{
		swap_int(&rect->pos.x, &rect->end.x);
	}

	if(rect->pos.y>rect->end.y)
	{
		swap_int(&rect->pos.y, &rect->end.y);
	}
}

int rect2_contains(const struct rect2* const rect, const struct vec2* const point)
{
	return
		/* horizontal */
		rect->pos.x<=point->x && point->x<=rect->end.x
		&&
		/* vertical */
		rect->pos.y<=point->y && point->y<=rect->end.y
		;
}
